using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using MacGruber;
using UnityEngine;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.Serialization;
using Object = UnityEngine.Object;
using Random = UnityEngine.Random;


namespace CheesyFX {

	public class ReadMyLips : MVRScript
	{
		public static ReadMyLips singleton;
		private UnityEventsListener uiListener;
		public static GenerateDAZMorphsControlUI morphControl;
		private static System.Random rng = new System.Random();

		private static JSONStorableBool voiceEnabledJ = new JSONStorableBool("Voice Enabled", true, val => voiceEnabled = val);
		private static JSONStorableBool expressionsEnabledJ = new JSONStorableBool("Expressions Enabled", true);
		private static JSONStorableBool breathingEnabledJ = new JSONStorableBool("Breathing Enabled", true, val => breathingDriver.enabled = val);
		// private JSONStorableBool voiceEnabled = new JSONStorableBool("Voice Enabled", true);
		private static List<ExpressionMorph> painMorphs = new List<ExpressionMorph>();
		private static List<ExpressionMorph> joyMorphs = new List<ExpressionMorph>();
		private static List<ExpressionMorph> lastMorphs = new List<ExpressionMorph>();
		private static JSONStorableBool autoJaw = new JSONStorableBool("Auto Jaw", true);
		private static JSONStorableBool displeaseOnly = new JSONStorableBool("Displeased Only", false);
		public static JSONStorableFloat painThresholdAtOne = CheesyUtils.SetupJSONFloat("Displease Threshold At High Stim", .1f, 0f, .5f, true);
		public static JSONStorableFloat painThresholdAtZero = CheesyUtils.SetupJSONFloat("Displease Threshold At Low Stim", 0f, 0f, .5f, true);
		public static JSONStorableFloat joyExpressionScale = CheesyUtils.SetupJSONFloat("Joy Expression Scale", 20f, 1f, 40f, false);
		public static JSONStorableFloat painExpressionScale = CheesyUtils.SetupJSONFloat("Pain Expression Scale", 10f, 1f, 40f, false);
		public static JSONStorableFloat idleThreshold = new JSONStorableFloat("Idle Threshold", .15f, 0f, 1f);
		public static JSONStorableFloat randomBurstChanceJ = new JSONStorableFloat("Random Burst Chance", .15f, 0f, 1f);
		public static JSONStorableFloat randomBurstStrenght = new JSONStorableFloat("Random Burst Strength", 7f, 0f, 10f);
		private static JSONStorableFloat maxLipLiftJ = new JSONStorableFloat("Max Upper Lip Lift", 4.5f, 0f, 10f);
		
		public static JSONStorableFloat jawLimit = new JSONStorableFloat("Jaw Limit", -12f, -35f, 0f);
		public static JSONStorableFloat headAudioVolume = new JSONStorableFloat("Head Audio Volume", 1f, 0f, 1f);
		public static JSONStorableFloat headAudioPitch = new JSONStorableFloat("Head Audio Pitch", 1f, -3f, 3f);
		
		private static JSONStorableFloat sexMoanChanceJ = new JSONStorableFloat("Sex Moan Chance", .1f, 0f, 1f, true);
		public static JSONStorableFloat slapMoanChance = new JSONStorableFloat("Slap Moan Chance", .6f, 0f, 1f, true);
		public static JSONStorableFloat sexSlapMoanChance = new JSONStorableFloat("Sex Slap Moan Chance", .4f, 0f, 1f, true);
		public static JSONStorableFloat burstMoanChance = new JSONStorableFloat("Burst Moan Chance", .5f, 0f, 1f, true);

		public static JSONStorableFloat orgasmCount = new JSONStorableFloat("Orgasms", 0f, 0f, 10f, false, false);
		public static JSONStorableFloat stimulation = new JSONStorableFloat("Stimulation", 0f, 0f, 1f, true, false);
		public static JSONStorableFloat stimulationGain = new JSONStorableFloat("Stimulation Gain", .1f, 0f, 1f, true);
		public static JSONStorableFloat stimulationRegressionJ = new JSONStorableFloat("Stimulation Regression", .15f, 0f, 1f, true);
		public static JSONStorableFloat orgasmThreshold = new JSONStorableFloat("Orgasm Threshold", 1f, 0f, 1f);
		public static JSONStorableFloat orgasmTime = new JSONStorableFloat("Orgasm Time", 10f, 0f, 30f, false);
		public static JSONStorableFloat orgasmFadeTime = new JSONStorableFloat("Orgasm Fade Time", 10f, 0f, 30f, false);
		public static JSONStorableFloat orgasmRecoverTime = new JSONStorableFloat("Orgasm Recover Time", 120f, 0f, 900f, false);
		public static JSONStorableAction orgasmNow = new JSONStorableAction("Orgasm Now", ForceOrgasm);
		public static JSONStorableAction resetOrgasms = new JSONStorableAction("Reset Orgasms", ResetOrgasms);
		
		private static float maxLipLift;
		private static float lipLiftSum;
		
		private static float lastStimulation;
		private static float stimRate;
		private FloatTriggerManager stimTriggers;
		private FloatTriggerManager orgasmTriggers;
		private static float resetTargetsTimer = 1f;
		private static float stimBurstTimer;
		public static float randomBurstChance;
		public static float sexMoanChance;
		public static float stimulationRegression;
		// private static float painThreshold;
		private static float orgasmTimer = 30f;
		private static float orgRecoverTimer = orgasmRecoverTime.defaultVal;
		private static bool isOrgasming;
		private static bool isIdling = true;
		
		private static float minJoyIncrement = .05f;
		private static float minPainIncrement = .05f;
		
		private bool vammoanPackageInstalled;
		public static List<ClipLibrary> moanLibrary;
		public static AudioSource headAudioSource;
		public static JSONStorableStringChooser vamMoanVoice;
		public static JSONStorableStringChooser voice = new JSONStorableStringChooser("vammoanSlapVoice",
			new List<string>{"Abby", "Claire", "Ella", "Emma", "Grace", "Isabella", "Lana", "Lia", "Lillian", "Lise", "Seth", "Skye", "Yumi"},
			"Isabella", "VAMMoan Voice");

		public static AudioSourceControl headAudioControl;
		public static DAZMeshEyelidControl eyelidBehavior;
		private static AdjustJoints jawControl;
		// private static JSONStorable autoJawMouthMorphs;
		public static float blinkTimeout;
		private List<object> UIElements = new List<object>();
		private UIDynamicTabBar tabbar;
		private int lastTabId;
		private UIDynamicButton stimTriggerButton;
		private UIDynamicButton orgasmTriggerButton;
		
		public static EventTrigger onOrgasmStartTrigger;
		public static EventTrigger onOrgasmEndTrigger;
		public static EventTrigger onOrgasmRecoverTrigger;

		private static bool voiceEnabled = true;
		private static bool expressionsEnabled = true;

		private PresetSystem presetSystem;
		private List<JSONStorableParam> parameters = new List<JSONStorableParam>();

		
		private static BreathingDriver breathingDriver;

		private static Ahegao ahegao;

		// private static EyesControl eyeBehavior;
		// private static DAZBone lEye;
		// private static DAZBone rEye;
		// private static EyesControl.LookMode lookMode;

		private JSONClass Store()
		{
			JSONClass jc = new JSONClass();
			parameters.ForEach(x => x.Store(jc));
			Ahegao.Store(jc);
			return jc;
		}

		private void Load(JSONClass jc)
		{
			parameters.ForEach(x => x.Load(jc));
			Ahegao.Load(jc);
		}

		private void OnEnable()
		{
			voiceEnabled = voiceEnabledJ.val;
			expressionsEnabled = expressionsEnabledJ.val;
		}

		public override void Init()
		{
			UIManager.SetScript(this, CreateUIElement, leftUIElements, rightUIElements);
			Utils.OnInitUI(CreateUIElement);
			singleton = this;
			// eyeBehavior = (EyesControl)containingAtom.GetStorableByID("Eyes");
			// lEye = containingAtom.GetComponentsInChildren<DAZBone>().First(x => x.name == "lEye");
			// rEye = containingAtom.GetComponentsInChildren<DAZBone>().First(x => x.name == "rEye");
			headAudioControl = (AudioSourceControl)containingAtom.GetStorableByID("HeadAudioSource");
			eyelidBehavior = (DAZMeshEyelidControl) containingAtom.GetStorableByID("EyelidControl");
			jawControl = (AdjustJoints)containingAtom.GetStorableByID("JawControl");
			jawControl.SetBoolParamValue("driveXRotationFromAudioSource", true);
			((SetDAZMorphControl)containingAtom.GetStorableByID("AutoExpressions")).enabledJSON.val = false;
			// autoJawMouthMorphs = containingAtom.GetStorableByID("AutoJawMouthMorph");
			DAZCharacterSelector characterSelector = containingAtom.GetStorableByID("geometry") as DAZCharacterSelector;
			morphControl = characterSelector.morphsControlUI;
			string joyMorphPath = TouchMe.packageUid + "Custom/Atom/Person/Morphs/female/CheesyFX/BodyLanguage/Expressions/Joy/";
			string painMorphPath = TouchMe.packageUid + "Custom/Atom/Person/Morphs/female/CheesyFX/BodyLanguage/Expressions/Displease";
			foreach (var path in FileManagerSecure.GetFiles(joyMorphPath + "Ascorad"))
			{
				if(path.EndsWith(".vmb")) continue;
				var pathNormalized = FileManagerSecure.NormalizePath(path);
				joyMorphs.Add(GetMorph(pathNormalized));
			}
			foreach (var path in FileManagerSecure.GetFiles(joyMorphPath + "Jackaroo"))
			{
				if(!path.EndsWith(".vmi")) continue;
				var pathNormalized = FileManagerSecure.NormalizePath(path);
				joyMorphs.Add(GetMorph(pathNormalized));
			}
			foreach (var path in FileManagerSecure.GetFiles(joyMorphPath + "AshAuryn"))
			{
				if(path.EndsWith(".vmb")) continue;
				var pathNormalized = FileManagerSecure.NormalizePath(path);
				joyMorphs.Add(GetMorph(pathNormalized));
			}
			foreach (var path in FileManagerSecure.GetFiles(joyMorphPath + "AshAuryn/AsymMouth"))
			{
				if(path.EndsWith(".vmb")) continue;
				var pathNormalized = FileManagerSecure.NormalizePath(path);
				joyMorphs.Add(GetMorph(pathNormalized));
			}
			foreach (var path in FileManagerSecure.GetFiles(painMorphPath))
			{
				if(path.EndsWith(".vmb")) continue;
				var pathNormalized = FileManagerSecure.NormalizePath(path);
				painMorphs.Add(GetMorph(pathNormalized));
			}
			
			joyMorphs.Add(GetMorph("Desire"));
			joyMorphs.Add(GetMorph("Excitement"));
			joyMorphs.Add(GetMorph("Flirting"));
			joyMorphs.Add(GetMorph("Flirting Feminine Left"));
			joyMorphs.Add(GetMorph("Flirting Feminine Right"));
			joyMorphs.Add(GetMorph("Flirting Masculine Left"));
			joyMorphs.Add(GetMorph("Flirting Masculine Right"));
			joyMorphs.Add(GetMorph("Frown"));
			joyMorphs.Add(GetMorph("Mouth Frown"));
			joyMorphs.Add(GetMorph("Sad"));
			joyMorphs.Add(GetMorph("Scream"));
			joyMorphs.Add(GetMorph("Shock"));
			joyMorphs.Add(GetMorph("Smile Full Face"));
			joyMorphs.Add(GetMorph("Snarl Left"));
			joyMorphs.Add(GetMorph("Snarl Right"));
			joyMorphs.Add(GetMorph("Surprise"));
			
			joyMorphs.Add(GetMorph("Brow Down"));
			joyMorphs.Add(GetMorph("Brow Down Left"));
			joyMorphs.Add(GetMorph("Brow Down Right"));
			joyMorphs.Add(GetMorph("Brow Inner Down"));
			joyMorphs.Add(GetMorph("Brow Inner Down Left"));
			joyMorphs.Add(GetMorph("Brow Inner Down Right"));
			joyMorphs.Add(GetMorph("Brow Inner Up Left"));
			joyMorphs.Add(GetMorph("Brow Inner Up Right"));
			joyMorphs.Add(GetMorph("Brow Outer Down Left"));
			joyMorphs.Add(GetMorph("Brow Outer Down Right"));
			joyMorphs.Add(GetMorph("Brow Outer Up"));
			joyMorphs.Add(GetMorph("Brow Outer Up Left"));
			joyMorphs.Add(GetMorph("Brow Outer Up Right"));
			joyMorphs.Add(GetMorph("Brow Up"));
			joyMorphs.Add(GetMorph("Brow Up Left"));
			joyMorphs.Add(GetMorph("Brow Up Right"));
			joyMorphs.Add(GetMorph("Brow Squeeze"));
			
			joyMorphs.Add(GetMorph("Eyes Closed"));
			joyMorphs.Add(GetMorph("Eyes Squint"));
			joyMorphs.Add(GetMorph("Eyes Squint Left"));
			joyMorphs.Add(GetMorph("Eyes Squint Right"));
			
			// joyMorphs.Add(GetMorph("Mouth Open"));
			// joyMorphs.Add(GetMorph("Mouth Open Wide"));
			// joyMorphs.Add(GetMorph("Mouth Narrow"));
			// joyMorphs.Add(GetMorph("Mouth Narrow Left"));
			// joyMorphs.Add(GetMorph("Mouth Narrow Right"));
			joyMorphs.Add(GetMorph("Mouth Side-Side Left"));
			joyMorphs.Add(GetMorph("Mouth Side-Side Right"));
			joyMorphs.Add(GetMorph("Mouth Smile"));
			joyMorphs.Add(GetMorph("Mouth Smile Open"));
			joyMorphs.Add(GetMorph("Mouth Smile Simple"));
			joyMorphs.Add(GetMorph("Mouth Smile Simple Left"));
			joyMorphs.Add(GetMorph("Mouth Smile Simple Right"));
			
			joyMorphs.Add(GetMorph("Lip Bottom Down"));
			joyMorphs.Add(GetMorph("Lip Bottom Out"));
			joyMorphs.Add(GetMorph("Lip Bottom In"));
			joyMorphs.Add(GetMorph("Lip Bottom In Left"));
			joyMorphs.Add(GetMorph("Lip Bottom Out Right"));
			joyMorphs.Add(GetMorph("Lip Bottom Up"));
			joyMorphs.Add(GetMorph("Lip Bottom Up Left"));
			joyMorphs.Add(GetMorph("Lip Bottom Up Right"));
			// joyMorphs.Add(GetMorph("Lip Top Up"));
			joyMorphs.Add(GetMorph("Lip Top Up Left"));
			joyMorphs.Add(GetMorph("Lip Top Up Right"));
			joyMorphs.Add(GetMorph("Lips Close"));
			joyMorphs.Add(GetMorph("Lips Part"));
			// joyMorphs.Add(GetMorph("Lips Pucker"));
			joyMorphs.Add(GetMorph("Lips Pucker Wide"));
			
			painMorphs.Add(GetMorph("Nose Wrinkle"));
			painMorphs.Add(GetMorph("Snarl Left"));
			painMorphs.Add(GetMorph("Snarl Right"));
			
			enabledJSON.setCallbackFunction += val => Reset();
			
			headAudioSource = containingAtom.GetStorableByID("HeadAudioSource").GetComponent<AudioSource>();
			moanLibrary = new List<ClipLibrary>
			{
				new ClipLibrary("moans0"),
				new ClipLibrary("moans1"),
				new ClipLibrary("moans2"),
				new ClipLibrary("moans3"),
				new ClipLibrary("moans4"),
				new ClipLibrary("moans5"),
			};
			var moansettings = new MoanSettings();
			
			var vammoan = this.FindPluginPeer("VAMMoan");
			if (vammoan)
			{
				vamMoanVoice = vammoan.GetStringChooserJSONParam("voice");
				voice.val = vamMoanVoice.val;
				vamMoanVoice.setCallbackFunction += val => voice.val = val;
			}
			else RetrySyncWithVamMoan().Start();
			voice.setCallbackFunction += val =>
			{
				StartCoroutine(AudioImporter.ReadBundle("VAMMoan/audio/voices.voicebundle", moanLibrary));
				if (vamMoanVoice != null) vamMoanVoice.val = val;
			};
			
			string vamMoanPackageUid = CheesyUtils.GetLatestPackageUid("hazmhox.vammoan");
			if (vamMoanPackageUid != null)
			{
				vammoanPackageInstalled = true;
				string moanPath = vamMoanPackageUid + ":/Custom/Scripts/VAMMoan/audio/voices.voicebundle";
				AudioImporter.GetClipsFromAssetBundle(moanLibrary, moanPath);
				var squishPath = vamMoanPackageUid + ":/Custom/Scripts/VAMMoan/audio/voices-shared.voicebundle";
				AudioImporter.GetClipsFromAssetBundle(new List<ClipLibrary>{FillMeUp.squishLibrary, FillMeUp.bjLibrary},
					squishPath);
			}
			else
			{
				vammoanPackageInstalled = false;
				slapMoanChance.val = 0f;
				sexSlapMoanChance.val = 0f;
			}
			slapMoanChance.setCallbackFunction += val =>
			{
				if (!vammoanPackageInstalled)
				{
					SuperController.LogError("SlapMe: Download VAMMoan to enable moan reactions.");
					slapMoanChance.valNoCallback = 0f;
				}
			};
			sexSlapMoanChance.setCallbackFunction += val =>
			{
				if (!vammoanPackageInstalled)
				{
					SuperController.LogError("SlapMe: Download VAMMoan to enable moan reactions.");
					sexSlapMoanChance.valNoCallback = 0f;
				}
			};
			randomBurstChanceJ.AddCallback(val => randomBurstChance = .01f * val);
			// idleBurstChanceJ.AddCallback(val => idleBurstChance = );
			sexMoanChanceJ.AddCallback(val => sexMoanChance = .1f * val);
			stimulationRegressionJ.AddCallback(val => stimulationRegression = val);
			orgasmRecoverTime.AddCallback(val => orgRecoverTimer = val);
			jawLimit.AddCallback((val => jawControl.SetFloatParamValue("driveXRotationFromAudioSourceMaxAngle", val)));
			headAudioVolume.AddCallback(val => headAudioControl.SetFloatParamValue("volume", val));
			headAudioPitch.AddCallback(val => headAudioControl.SetFloatParamValue("pitch", val));
			expressionsEnabledJ.AddCallback(val =>
			{
				expressionsEnabled = val;
				if(!val) ResetExpressions();
			});
			joyExpressionScale.AddCallback(val =>
			{
				if (val <= 20f) minJoyIncrement = .05f;
				else minJoyIncrement = .0025f * val;
			});
			painExpressionScale.AddCallback(val =>
			{
				if (val <= 20f) minPainIncrement = .05f;
				else minPainIncrement = .0025f * val;
			});
			stimulation.setCallbackFunction += val =>
			{
				if (isIdling)
				{
					if (val > idleThreshold.val) SetIdleMode(false);
				}
				else if (val < idleThreshold.val) SetIdleMode(true);
			};
			maxLipLiftJ.AddCallback(val => maxLipLift = .001f * val);
			
			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			
			lastMorphs.Add(joyMorphs[0]);

			onOrgasmStartTrigger = new EventTrigger(this, "On Orgasm Start");
			onOrgasmEndTrigger = new EventTrigger(this, "On Orgasm End");
			onOrgasmRecoverTrigger = new EventTrigger(this, "On Orgasm Recover");
			stimTriggers = this.AddFloatTriggerManager(stimulation);
			orgasmTriggers = this.AddFloatTriggerManager(orgasmCount);
			
			var labiaListener = BodyRegionMapping.touchZones["Labia"].touchCollisionListener;
			
			parameters.Add(randomBurstChanceJ);
			parameters.Add(randomBurstStrenght);
			parameters.Add(joyExpressionScale);
			parameters.Add(painThresholdAtZero);
			parameters.Add(painThresholdAtOne);
			parameters.Add(painExpressionScale);
			parameters.Add(jawLimit);
			parameters.Add(headAudioVolume);
			parameters.Add(headAudioPitch);
			parameters.Add(idleThreshold);
			parameters.Add(stimulationGain);
			parameters.Add(stimulationRegressionJ);
			parameters.Add(orgasmThreshold);
			parameters.Add(orgasmTime);
			parameters.Add(orgasmFadeTime);
			parameters.Add(orgasmRecoverTime);
			parameters.Add(sexMoanChanceJ);
			parameters.Add(slapMoanChance);
			parameters.Add(sexSlapMoanChance);
			parameters.Add(burstMoanChance);
			parameters.Add(voice);
			parameters.Add(expressionsEnabledJ);
			parameters.Add(voiceEnabledJ);
			parameters.Add(breathingEnabledJ);
			parameters.Add(maxLipLiftJ);
			parameters.Add(displeaseOnly);
			parameters.AddRange(MoanSettings.thresholds.Select(x => x as JSONStorableParam));
			
			parameters.ForEach(x => x.Register(this));
			RegisterAction(orgasmNow);

			presetSystem = new PresetSystem(this)
			{
				saveDir = "Saves/PluginData/CheesyFX/BodyLanguage/ReadMyLips/",
				Store = Store,
				Load = Load
			};
			presetSystem.Init();
			CreateUI();
			SelectTab(1);
			SelectTab(0);
			Reset();

			breathingDriver = gameObject.AddComponent<BreathingDriver>().Init();
			stimulation.setCallbackFunction += breathingDriver.SetParameters;
			
			ahegao = gameObject.AddComponent<Ahegao>().Init();

			// PrintMorphFile();
			LoadExpressionSets();
		}

		public override void InitUI()
		{
			base.InitUI();
			if(UITransform == null || uiListener != null) return;
			uiListener = UITransform.gameObject.AddComponent<UnityEventsListener>();
			uiListener.onEnabled.AddListener(() => UIManager.SetScript(this, CreateUIElement, leftUIElements, rightUIElements));
			uiListener.onEnabled.AddListener(() => Utils.OnInitUI(CreateUIElement));
		}

		private void OnDisable()
		{
			voiceEnabled = false;
			expressionsEnabled = false;
		}

		private void OnDestroy()
		{
			Reset();
			onOrgasmStartTrigger.Remove();
			onOrgasmEndTrigger.Remove();
			onOrgasmRecoverTrigger.Remove();
			Destroy(uiListener);
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[onOrgasmStartTrigger.Name] = onOrgasmStartTrigger.GetJSON(base.subScenePrefix);
				jc[onOrgasmEndTrigger.Name] = onOrgasmEndTrigger.GetJSON(base.subScenePrefix);
				jc[onOrgasmRecoverTrigger.Name] = onOrgasmRecoverTrigger.GetJSON(base.subScenePrefix);
				jc["StimTriggers"] = stimTriggers.Store();
				jc["OrgasmTriggers"] = orgasmTriggers.Store();
			}
			Ahegao.Store(jc);
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			Load(jc);
			if (!physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				needsStore = true;
				
				onOrgasmStartTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				onOrgasmEndTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				onOrgasmRecoverTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				
				stimTriggers.Load(jc["StimTriggers"].AsObject);
				orgasmTriggers.Load(jc["OrgasmTriggers"].AsObject);
				stimTriggerButton.label = $"Stimulation Triggers ({stimTriggers.triggers.Count})";
				orgasmTriggerButton.label = $"Orgasm Count Triggers ({orgasmTriggers.triggers.Count})";
				
				foreach (var setting in MoanSettings.thresholds)
				{
					setting.RestoreFromJSON(jc);
				}
			}
		}

		public void OnAtomRename(string oldUid, string newUid)
		{
			// stimTriggers.ForEach(x => x.OnAtomRename());
			// orgasmTriggers.ForEach(x => x.OnAtomRename());
			onOrgasmStartTrigger.SyncAtomNames();
			onOrgasmEndTrigger.SyncAtomNames();
			onOrgasmRecoverTrigger.SyncAtomNames();
		}
		
		private void CreateUI()
		{
			presetSystem.CreateUI();
			// UIManager.SetScript(this, CreateUIElement, leftUIElements, rightUIElements);
			UIManager.RemoveUIElements(UIElements);
			stimulation.CreateUI(this);
			orgasmCount.CreateUI(this, true);
			tabbar = UIManager.CreateTabBar(new [] {"Stimulation", "Orgasm", "Ahegao", "Expressions", "Voice"}, SelectTab, script:this);
			tabbar.SelectTab(lastTabId);
		}
		
		public void ClearUI()
		{
			UIManager.RemoveUIElements(leftUIElements.Select(x => (object) x.GetComponent<UIDynamic>()).ToList());
			UIManager.RemoveUIElements(rightUIElements.Select(x => (object) x.GetComponent<UIDynamic>()).ToList());
		}

		private void SelectTab(int id)
		{
			lastTabId = id;
			UIManager.RemoveUIElements(UIElements);
			if(id == 0) CreateStimUI();
			else if (id == 1) CreateOrgasmUI();
			else if (id == 2) Ahegao.CreateUI(UIElements);
			else if (id == 3) CreateExpressionsUI();
			else CreateVoiceUI();
		}

		public static JSONStorableFloat testStim = new JSONStorableFloat("Test Stim", 0f, 0f, 1f);

		private void CreateStimUI()
		{
			stimulationGain.CreateUI(script: this, rightSide:false, UIElements:UIElements);
			stimulationRegressionJ.CreateUI(script: this, rightSide:true, UIElements:UIElements);
			stimTriggerButton = CreateButton($"Stimulation Triggers ({stimTriggers.triggers.Count})", false);
			stimTriggerButton.buttonColor = new Color(0.45f, 1f, 0.45f);
			stimTriggerButton.button.onClick.AddListener(() =>
			{
				{
					ClearUI();
					stimTriggers.OpenPanel(this, CreateUI);
				}
			});
			UIElements.Add(stimTriggerButton);
			breathingEnabledJ.CreateUI(script:this, rightSide:true, UIElements:UIElements);
			// testStim.CreateUI(script:this, UIElements:UIElements);
		}

		private void CreateOrgasmUI()
		{
			orgasmThreshold.CreateUI(UIElements: UIElements);
			orgasmTime.CreateUI(UIElements: UIElements);
			orgasmFadeTime.CreateUI(UIElements: UIElements);
			orgasmRecoverTime.CreateUI(UIElements: UIElements);
			
			orgasmTriggerButton = CreateButton($"Orgasm Count Triggers ({orgasmTriggers.triggers.Count})", true);
			orgasmTriggerButton.buttonColor = new Color(0.45f, 1f, 0.45f);
			orgasmTriggerButton.button.onClick.AddListener(() =>
			{
				{
					ClearUI();
					// orgasmTriggers.triggers[0].OpenPanel(this, CreateUI);
					orgasmTriggers.OpenPanel(this, CreateUI);
				}
			});
			UIElements.Add(orgasmTriggerButton);
			var button = CreateButton("On Orgasm Start", true);
			button.buttonColor = new Color(0.45f, 1f, 0.45f);
			button.button.onClick.AddListener(onOrgasmStartTrigger.OpenPanel);
			UIElements.Add(button);
			button = CreateButton("On Orgasm End", true);
			button.buttonColor = new Color(0.45f, 1f, 0.45f);
			button.button.onClick.AddListener(onOrgasmEndTrigger.OpenPanel);
			UIElements.Add(button);
			button = CreateButton("On Orgasm Recover", true);
			button.buttonColor = new Color(0.45f, 1f, 0.45f);
			button.button.onClick.AddListener(onOrgasmRecoverTrigger.OpenPanel);
			UIElements.Add(button);
			button = CreateButton("Orgasm Now", true);
			button.button.onClick.AddListener(ForceOrgasm);
			UIElements.Add(button);
			button = CreateButton("Reset Orgasms", true);
			button.button.onClick.AddListener(resetOrgasms.actionCallback.Invoke);
			UIElements.Add(button);
		}

		private void CreateExpressionsUI()
		{
			joyExpressionScale.CreateUI(this, UIElements:UIElements);
			painExpressionScale.CreateUI(this, UIElements:UIElements, rightSide:true);
			// maxMorphVal.CreateUI(this, UIElements:UIElements);
			painThresholdAtZero.CreateUI(UIElements: UIElements, rightSide:true);
			painThresholdAtOne.CreateUI(UIElements: UIElements, rightSide:true);
			randomBurstChanceJ.CreateUI(this, UIElements:UIElements);
			randomBurstStrenght.CreateUI(this, UIElements:UIElements);
			idleThreshold.CreateUI(this, UIElements:UIElements, rightSide:false);
			maxLipLiftJ.CreateUI(this, UIElements:UIElements, rightSide:true);
			expressionsEnabledJ.CreateUI(this, UIElements: UIElements);
			displeaseOnly.CreateUI(this, UIElements: UIElements, rightSide:true);
		}
		
		private void CreateVoiceUI()
		{
			voiceEnabledJ.CreateUI(this, UIElements: UIElements);
			voice.CreateUI(script: this, rightSide:true, chooserType:1, UIElements:UIElements);
			sexMoanChanceJ.CreateUI(script: this, UIElements:UIElements, rightSide:true);
			burstMoanChance.CreateUI(script: this, UIElements:UIElements, rightSide:true);
			slapMoanChance.CreateUI(script: this, UIElements:UIElements, rightSide:true);
			sexSlapMoanChance.CreateUI(script: this, rightSide:true, UIElements:UIElements);
			
			foreach (var setting in MoanSettings.thresholds)
			{
				setting.CreateUI(script: this, UIElements: UIElements);
			}
			orgasmThreshold.CreateUI(script: this, UIElements: UIElements);
			headAudioVolume.CreateUI(script: this, UIElements:UIElements, rightSide:true);
			headAudioPitch.CreateUI(script: this, UIElements:UIElements, rightSide:true);
			jawLimit.CreateUI(this, UIElements:UIElements, rightSide:false);
		}

		public static void Stimulate(float stimulus, bool doChange = false, bool doStim = false)
		{
			if (SuperController.singleton.freezeAnimation) return;
			float stim = stimulus * .01f;
			stimBurstTimer = .2f;
			if (doChange)
			{
				if (stimulation.val <= .01f)
				{
					resetTargetsTimer = stimBurstTimer = Random.Range(.5f, 4f);
				}
				resetTargetsTimer = 1f;
			}
			DistributeStimulus(stim*(1f+2f*stimulation.val*stimulation.val), doChange);
			if(FillMeUp.isPenetrated || doStim)
			{
				stimulation.val += stim * stimulationGain.val;
				if (!isOrgasmPleasure && stimulation.val >= orgasmThreshold.val)
				{
					DoOrgasm();
					// isOrgasming = true;
					// isOrgasmPleasure = true;
					// // isOrgasmFade = true;
					// orgasmCount.val += 1f;
					// Stimulate(3f, true);
					// onOrgasmStartTrigger.Trigger();
					// ToggleMaleOrgasm(true);
					// shakers.ForEach(x => x.enabled = true);
					// float eyeRoll = Random.Range(-2f, -1f);
					// orgasmMorphs.ForEach(x => x.target = eyeRoll);
					// eyeRoll.Print();
				}
			}
		}

		public static void DistributeStimulus(float stim, bool doChange)
		{
			if (!expressionsEnabled) return;
			float available = stim;
			List<ExpressionMorph> morphs;
			float factor;
			float minIncrement;
			if(displeaseOnly.val || (stimRate > 0f && stimRate >= Mathf.Lerp(painThresholdAtZero.val, painThresholdAtOne.val, stimulation.val))){
				morphs = new List<ExpressionMorph>(painMorphs);
				factor = painExpressionScale.val;
				minIncrement = minPainIncrement;
			}
			else{
				morphs = new List<ExpressionMorph>(joyMorphs);
				factor = Mathf.Lerp(joyExpressionScale.val*2f, joyExpressionScale.val * .5f, stimulation.val);
				// factor = joyExpressionScale.val * 1.5f;
				minIncrement = minJoyIncrement;
			}

			GetLipLiftSum();
			available *= factor;
			if (!doChange && Random.Range(0f, 1f) > .1f)
			{
				foreach(var morph in lastMorphs)
				{
					if (available <= minIncrement) break;
					float maxIncrement = Math.Min(available, morph.max - morph.target);
					if(morph.lipLift > 0f) maxIncrement = Mathf.Min(maxIncrement, (maxLipLift-lipLiftSum)/morph.lipLift);
					if(maxIncrement < minIncrement) continue;
					float increment = Random.Range(minIncrement, maxIncrement);
					morph.target += increment;
					available -= increment;
					lipLiftSum += morph.target * morph.lipLift;
				}
			}
			else
			{
				lastMorphs.Clear();
				for (var i = 0; i < morphs.Count; i++)
				{
					if (available <= minIncrement) break;
					var morph = morphs[rng.Next(morphs.Count)];
					float maxIncrement = Math.Min(available, morph.max - morph.target);
					if(morph.lipLift > 0f) maxIncrement = Mathf.Min(maxIncrement, (maxLipLift-lipLiftSum)/morph.lipLift);
					if(maxIncrement < minIncrement) continue;
					float increment = Random.Range(minIncrement, maxIncrement);
					morph.target += increment;
					available -= increment;
					lastMorphs.Add(morph);
					lipLiftSum += morph.target * morph.lipLift;
				}
			}
			if(targetsReset) targetsReset = false;
			// if(lipLiftSum > .0025f) lipLiftSum.Print();
			// lipLiftSum = 0f;
		}
		
		private static bool targetsReset;
		private float burstTimer;
		private static bool isOrgasmPleasure;

		private static void GetLipLiftSum()
		{
			lipLiftSum = 0f;
			for (int i = 0; i < joyMorphs.Count; i++)
			{
				if(!joyMorphs[i].active) continue;
				lipLiftSum += joyMorphs[i].lipLift * joyMorphs[i].target;
			}
			for (int i = 0; i < painMorphs.Count; i++)
			{
				if(!painMorphs[i].active) continue;
				lipLiftSum += painMorphs[i].lipLift * painMorphs[i].target;
			}
			// lipLiftSum.Print();
		}
		
		public void Update()
		{
			if (SuperController.singleton.freezeAnimation) return;
			// lEye.transform.localEulerAngles = new Vector3(-25f, 0f, 0f);
			// GetLipLiftSum();
			onOrgasmStartTrigger.Update();
			onOrgasmEndTrigger.Update();
			onOrgasmRecoverTrigger.Update();
			stimRate = Mathf.Max((stimulation.val - lastStimulation) / Time.deltaTime, 0f);
			lastStimulation = stimulation.val;
			// stimRate.Print();
			// if(stimRate > .1f) stimRate.Print();
			if(isIdling)
			{
				burstTimer -= Time.deltaTime;
				if (burstTimer < 0f)
				{
					burstTimer = .2f;

					if (randomBurstChanceJ.val >= .999f || (randomBurstChance > 0f && randomBurstStrenght.val > 0f &&
					    Random.Range(0f, 1f) <= randomBurstChanceJ.val))
					{
						Stimulate(randomBurstStrenght.val, true);
						PlayBurstMoan();
					}
				}
			}
			
			if (isOrgasming)
			{
				orgasmTimer -= Time.deltaTime;
				if (orgasmTimer < orgasmFadeTime.val)
				{
					// painThreshold = stimulation.val - .05f;
					if(isOrgasmPleasure)
					{
						isOrgasmPleasure = false;
						stimulationRegression = stimulationRegressionJ.val + .1f*orgasmCount.val;
						Ahegao.ShutDown();
						// eyeControl.lEyeTarget = eyeControl.rEyeTarget = Quaternion.identity;
						// eyeControl.lEyeTarget = eyeControl.rEyeTarget = eyeControl.GetLookAtRotation();
						// eyeBehavior.currentLookMode = lookMode;
						// eyeControl.enabled = false;
					}

					// if (ahegao.enabled && orgasmTimer < Mathf.Max(orgasmFadeTime - 2f, 0f))
					// 	ahegao.enabled = false;
					if (orgasmTimer < 0f)
					{
						onOrgasmEndTrigger.Trigger();
						if (orgasmCount.val > 0)
							stimulationRegression =
								stimulationRegressionJ.val * Mathf.Log(2f * orgasmCount.val + 1f);
						else stimulationRegression = stimulationRegressionJ.val;
						isOrgasming = false;
						// ahegao.enabled = false;
					}
				}
			}
			else if(orgasmCount.val > 0)
			{
				orgRecoverTimer -= Time.deltaTime;
				if (orgRecoverTimer < 0f)
				{
					onOrgasmRecoverTrigger.Trigger();
					if (orgasmCount.val > 0)
					{
						orgasmCount.val -= 1;
						stimulationRegression = stimulationRegressionJ.val * Mathf.Log(2f * orgasmCount.val + 1f);
					}
					orgRecoverTimer = orgasmRecoverTime.val;
				}
			}
			
			if(blinkTimeout > 0f) blinkTimeout -= Time.deltaTime;
			if(stimulation.val > .01f){
				stimulation.val = Mathf.Lerp(stimulation.val, 0f, stimulationRegression*Time.deltaTime);
			}
			
			// if(lastMorphs[0].target >0f) lastMorphs[0].target.Print();
			// else if(!isReset) ResetTargets();
			if(resetTargetsTimer > 0f) resetTargetsTimer -= Time.deltaTime;
			if(stimBurstTimer > 0f) stimBurstTimer -= Time.deltaTime;
			
			foreach(var morph in painMorphs)
			{
				morph.LerpToTarget();
			}
			foreach(var morph in joyMorphs){
				// if(morph.target > 0f) $"{morph.name} {morph.target} {morph.val}".Print();
				morph.LerpToTarget();
			}
			if(!targetsReset && (resetTargetsTimer <= 0f || stimBurstTimer <= 0f))
			{
				foreach(var morph in painMorphs){
					if(morph.target > 0f) morph.target = 0f;
				}
				foreach(var morph in joyMorphs){
					if(morph.target > 0f) morph.target = 0f;
				}
				targetsReset = true;
				resetTargetsTimer = 1f;
			}
		}

		private static void DoOrgasm()
		{
			// lookMode = eyeBehavior.currentLookMode;
			// eyeBehavior.currentLookMode = EyesControl.LookMode.None;
			isOrgasming = true;
			isOrgasmPleasure = true;
			orgasmCount.val += 1f;
			Stimulate(3f, true);
			onOrgasmStartTrigger.Trigger();
			// float eyeRoll = Random.Range(-22, -10f);
			ahegao.Run();
			orgasmTimer = orgasmTime.val + orgasmFadeTime.val;
			PlayRandomMoan(5);
			// orgasmMorphs.ForEach(x => x.target = eyeRoll);
			// lEye.transform.localEulerAngles = new Vector3(eyeRoll, 12f, 0f);
			// rEye.transform.localEulerAngles = new Vector3(eyeRoll, -12f, 0f);
		}

		public static void ForceOrgasm()
		{
			orgasmTimer = orgasmTime.val;
			stimulation.val = 1f;
			stimulationRegression = stimulationRegressionJ.val;
			DoOrgasm();
		}
		
		private static void ResetOrgasms()
		{
			orgasmCount.val = 0f;
			// orgRecoverTimer = orgasmRecoverTime.val;
			stimulationRegression = stimulationRegressionJ.val;
			// orgasmTimer = orgasmTime.val;
		}

		private static void ResetExpressions()
		{
			foreach(var morph in painMorphs)
			{
				morph.val = 0f;
				morph.target = 0f;
			}
			foreach (var morph in joyMorphs)
			{
				morph.val = 0f;
				morph.target = 0f;
			}
		}

		public void Reset()
		{
			ResetExpressions();
			if (isOrgasming) ahegao.OnDisable();
			stimulation.val = 0f;
		}

		public static ExpressionMorph GetMorph(string uid)
		{
			DAZMorph morph = morphControl.GetMorphByUid(uid);
			if (morph == null) uid.Print();
			// morph.displayName.Print();
			// morph.morphValue.Print();
			return new ExpressionMorph(morph);
		}

		public static void PlaySlapMoan(bool isSexSlap = false)
		{
			if(!voiceEnabled || FillMeUp.throat.depth.val > .02f || headAudioSource.isPlaying) return;
			if (isSexSlap)
			{
				if (sexSlapMoanChance.val > 0f)
				{
					if(Random.Range(0f, 1f) < sexSlapMoanChance.val) PlayRandomMoan(GetMoanLevel());
				}
			}
			else if(slapMoanChance.val > 0f && Random.Range(0f, 1f) <= slapMoanChance.val) PlayRandomMoan(GetMoanLevel());
		}

		private static void PlayRandomMoan(int minLvl)
		{
			moanLibrary[rng.Next(minLvl, 5)].Play(headAudioSource);
		}
		
		public static void PlayBurstMoan()
		{
			if(!voiceEnabled || FillMeUp.throat.depth.val > .02f || headAudioSource.isPlaying || burstMoanChance.val == 0f) return;
			if (Random.Range(0f, 1f) <= burstMoanChance.val)
			{
				moanLibrary[GetMoanLevel()].Play(headAudioSource);
			}
		}
		
		public static bool PlayMoan()
		{
			if(!voiceEnabled || FillMeUp.throat.depth.val > .02f || headAudioSource.isPlaying || sexMoanChance == 0f || stimulation.val == 0f) return false;
			if (Random.Range(0f, 1f) <= sexMoanChance * stimulation.val)
			{
				moanLibrary[GetMoanLevel()].Play(headAudioSource);
				return true;
			}
			return false;
		}

		private static int GetMoanLevel()
		{
			if (isOrgasmPleasure) return 5;
			if(stimulation.val < MoanSettings.thresholds[0].val) return 0;
			for (var i = MoanSettings.thresholds.Count - 1; i > 0; i--)
			{
				if (stimulation.val <= MoanSettings.thresholds[i].val) continue;
				return i+1;
			}
			return 1;
		}
		
		private IEnumerator RetrySyncWithVamMoan()
		{
			yield return new WaitForEndOfFrame();
			for (int i = 0; i < 5; i++)
			{
				yield return new WaitForSeconds(2f);
				var vammoan = this.FindPluginPeer("VAMMoan");
				if (vammoan)
				{
					vamMoanVoice = vammoan.GetStringChooserJSONParam("voice");
					voice.val = vamMoanVoice.val;
					vamMoanVoice.setCallbackFunction += val => voice.val = val;
					yield break;
				}
			}
		}

		// public void AddTrigger()
		// {
		// 	var trigger = gameObject.AddComponent<CustomFloatTrigger>();
		// 	trigger.Init(this, stimulation, $"{stimulation.name}_{stimTriggers.Count}");
		// 	stimTriggers.Add(trigger);
		// 	trigger = gameObject.AddComponent<CustomFloatTrigger>();
		// 	trigger.Init(this, orgasmCount, $"{orgasmCount.name}_{orgasmTriggers.Count}", false, 1f);
		// 	orgasmTriggers.Add(trigger);
		// }

		private void PrintMorphFile()
		{
			FileManagerSecure.CreateDirectory("Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/joy");
			FileManagerSecure.CreateDirectory("Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/displease");
			JSONClass existing = null;
			JSONClass jc = new JSONClass();
			if(FileManagerSecure.FileExists("Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/joy/general.json"))
			{
				jc = LoadJSON("Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/joy/general.json").AsObject;
				// FileManagerSecure.DeleteFile("Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/joy/general.json");
			}
			foreach (var morph in joyMorphs)
			{
				if(jc.HasKey(morph.uid)) continue;
				morph.dazMorph.LoadDeltas();
				var vertex = morph.dazMorph.deltas.FirstOrDefault(x => x.vertex == 2138);
				// if(vertex != null) $"{morph.uid}: {vertex.delta.x}, {vertex.delta.y}, {vertex.delta.z}".Print();
				jc["set"] = "general";
				// jc[morph.uid]["isSmile"].AsBool = morph.name.ToLower().Contains("smile") || (existing != null && existing[morph.uid].AsObject["isSmile"].AsBool);
				jc[morph.uid]["lipLift"].AsFloat = vertex == null ? 0f : vertex.delta.y;
				jc[morph.uid]["weight"].AsFloat = 1f;
				jc[morph.uid]["max"].AsFloat = 0.1f;
				jc[morph.uid]["exclude"] = new JSONArray();
			}
			SaveJSON(jc, "Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/joy/general.json");
			
			if(FileManagerSecure.FileExists("Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/displease/general.json"))
			{
				jc = LoadJSON("Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/displease/general.json").AsObject;
				// FileManagerSecure.DeleteFile("Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/displease/general.json");
			}
			else jc = new JSONClass();
			foreach (var morph in painMorphs)
			{
				if(jc.HasKey(morph.uid)) continue;
				morph.dazMorph.LoadDeltas();
				var vertex = morph.dazMorph.deltas.FirstOrDefault(x => x.vertex == 2138);
				jc["set"] = "general";
				// jc[morph.uid]["isSmile"].AsBool = morph.name.ToLower().Contains("smile") || (existing != null && existing[morph.uid].AsObject["isSmile"].AsBool);
				jc[morph.uid]["lipLift"].AsFloat = vertex == null ? 0f : vertex.delta.y;
				jc[morph.uid]["weight"].AsFloat = 1f;
				jc[morph.uid]["max"].AsFloat = 0.1f;
				jc[morph.uid]["exclude"] = new JSONArray();
			}
			SaveJSON(jc, "Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/displease/general.json");
		}

		private Dictionary<string, JSONClass> expressionSets = new Dictionary<string, JSONClass>();
		private void LoadExpressionSets()
		{
			string packageUid = Utils.GetPackagePath(this);
			var joySets = packageUid + "Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/joy";
			var painSets = packageUid + "Custom/Scripts/CheesyFX/BodyLanguage/ExpressionSets/displease";
			JSONClass jc;
			JSONClass morphJc;
			foreach (var path in FileManagerSecure.GetFiles(joySets))
			{
				var normalizedPath = FileManagerSecure.NormalizePath(path);
				jc = LoadJSON(normalizedPath).AsObject;
				expressionSets[jc["set"]] = jc;
			}
			jc = expressionSets["general"];
			foreach (var morph in joyMorphs)
			{
				if(!jc.HasKey(morph.uid)) continue;
				morphJc = jc[morph.uid].AsObject;
				morph.lipLift = morphJc["lipLift"].AsFloat;
				morph.max = morphJc["max"].AsFloat;
			}
			foreach (var path in FileManagerSecure.GetFiles(painSets))
			{
				var normalizedPath = FileManagerSecure.NormalizePath(path);
				jc = LoadJSON(normalizedPath).AsObject;
				expressionSets[jc["set"]] = jc;
			}
			jc = expressionSets["general"];
			foreach (var morph in painMorphs)
			{
				if(!jc.HasKey(morph.uid)) continue;
				morphJc = jc[morph.uid].AsObject;
				morph.lipLift = morphJc["lipLift"].AsFloat;
				morph.max = morphJc["max"].AsFloat;
			}
		}

		public static void SetIdleMode(bool val)
		{
			// var idling = stimulation.val < .05f || !(FillMeUp.orifices.Any(x => x.isPenetrated) || BodyRegionMapping.touchZones["Labia"].touchCollisionListener.isOnStay);
			if (isIdling == val) return;
			isIdling = val;
			if (val)
			{
				ExpressionMorph.quicknessIn = 2f;
				ExpressionMorph.quicknessOut = 1f;
			}
			else
			{
				ExpressionMorph.quicknessIn = 4f;
				ExpressionMorph.quicknessOut = 2f;
			}
		}
	}

	public class ExpressionMorph
	{
		public DAZMorph dazMorph;
		public float lipLift;
		public float weight;
		public float max;
		private bool isAtTarget = true;
		public bool active;
		public static float quicknessIn = 2f;
		public static float quicknessOut = .5f;
		public string name => dazMorph.displayName;
		public string uid => dazMorph.uid;
		private float _target;
		public float target
		{
			get { return _target; }
			set
			{
				if (_target == value) return;
				_target = value;
				isAtTarget = false;
				if (value > 0f) active = true;
			}
		}

		public float val
		{
			get { return dazMorph.morphValue; }
			set { dazMorph.morphValue = value; }
		}

		public ExpressionMorph(DAZMorph dazMorph)
		{
			this.dazMorph = dazMorph;
		}

		public void LerpToTarget()
		{
			if (isAtTarget) return;
			if (!isAtTarget && Mathf.Abs(val - _target) < .001f)
			{
				val = _target;
				isAtTarget = true;
				if (_target == 0f) active = false;
				// if(uid.Contains("Eyeball")) val.Print();
				return;
			}
			if(_target > 0) val = Mathf.Lerp(val, _target, quicknessIn*Time.deltaTime);
			else val = Mathf.Lerp(val, _target, quicknessOut*Time.deltaTime);
		}
	}

	public class MoanSettings
	{
		public static List<JSONStorableFloat> thresholds = new List<JSONStorableFloat>();

		public MoanSettings()
		{
			for (int i = 0; i < 4; i++)
			{
				var setting = new JSONStorableFloat($"Level {i+1} Threshold", .1f + .2f * i, 0f, 1f);
				thresholds.Add(setting);
				// setting.setCallbackFunction += val => SyncRanges();
			}
		}

		// private void SyncRanges()
		// {
		// 	for (var i = 0; i < thresholds.Count-1; i++)
		// 	{
		// 		thresholds[i + 1].min = thresholds[i].val;
		// 		thresholds[i].max = thresholds[i+1].val;
		// 	}
		// }
	}
}