using System;
using UnityEngine;
using Random = UnityEngine.Random;

namespace CheesyFX
{
    public class Shaker : MonoBehaviour
    {
        protected Rigidbody RB;
        protected int axis;
        private float timer;
        private float forceTimer;
        private float flip;
        private bool applyForceOnReturn = true;

        private Vector3 currentForce;
        private Vector3 currentTorque;
        protected Vector3 targetForce;
        private Vector3 targetTorque;
        public float forceFactor;
        protected float randomForceFactor;
        protected float rotation;

        protected float periodRatio = .5f;
        protected float period = .15f;
        public float baseForceFactor = 400f;
        private float forceDuration = 1f;
        protected float forceQuickness = 1f;
        private float torqueQuickness = 1f;

        private float rng;
        private int rngCounter;
        private bool shutDown;
        private float shutDownFactor;

        private bool ready;

        public Shaker Init(Rigidbody rb, int axis = 0)
        {
            RB = rb;
            this.axis = axis;
            if(RB.name.StartsWith("l")) baseForceFactor = -baseForceFactor;
            if (axis == 2)
            {
                baseForceFactor *= 3f;
            }

            forceFactor = baseForceFactor;
            enabled = false;
            ready = true;
            return this;
        }

        public virtual void Randomize()
        {
            forceQuickness = Random.Range(1f, 2f);
            randomForceFactor = Random.Range(forceFactor*.8f, forceFactor*1.5f);
            switch (axis)
            {
                case 0:
                    period = Random.Range(.15f, .25f);
                    periodRatio = Random.Range(.3f, .8f);
                    rotation = Random.Range(0f, 90f);
                    break;
                case 2:
                    period = Random.Range(.2f, .3f);
                    periodRatio = Random.Range(.2f, .7f);
                    rotation = Random.Range(0f, 45f);
                    break;
            }
        }

        private void SetTargets()
        {
            timer -= Time.fixedDeltaTime;
            forceTimer -= Time.fixedDeltaTime;
            if (Random.Range(0f, 1f) < .1f)
            {
                targetForce = Vector3.zero;
                return;
            }
            if (timer < 0.0f) {
                if ((flip > 0f && periodRatio != 1f) || periodRatio == 0f) {
                    if (applyForceOnReturn) {
                        flip = -1f;
                    } else {
                        flip = 0f;
                    }
                    timer = period * (1f - periodRatio);
                    forceTimer = forceDuration * period;
                } else {
                    flip = 1f;
                    timer = period * periodRatio;
                    forceTimer = forceDuration * period;
                }
                SetForce(flip);
            } else if (forceTimer < 0.0f) {
                SetForce(0f);
            }
        }

        void FixedUpdate() {
            try
            {
                if (SuperController.singleton.freezeAnimation) return;
                rngCounter++;
                if (rngCounter > 30)
                {
                    Randomize();
                    rngCounter = 0;
                }
                SetTargets();
                if (shutDown)
                {
                    shutDownFactor -= Time.fixedDeltaTime;
                    targetForce *= .25f * shutDownFactor;
                    if (shutDownFactor < 0f)
                    {
                        enabled = false;
                        shutDownFactor = 4f;
                        shutDown = false;
                    }
                }
                currentForce = Vector3.Lerp(currentForce, targetForce, Time.fixedDeltaTime * forceQuickness);
                RB.AddForce(currentForce, ForceMode.Force);
            }
            catch (Exception e) {
                SuperController.LogError("Exception caught: " + e);
            }
        }
        
        public virtual void SetForce(float percent) {
            switch (axis)
            {
                case 0:
                    targetForce = percent * randomForceFactor * (Quaternion.AngleAxis(rotation, RB.transform.up) * RB.transform.right);
                    break;
                case 1:
                    targetForce = percent * randomForceFactor * RB.transform.up;
                    break;
                case 2:
                    targetForce = percent * randomForceFactor * (Quaternion.AngleAxis(rotation, RB.transform.right) * RB.transform.forward);
                    break;
            }
        }

        public void Start(float strengt)
        {
            forceFactor = baseForceFactor * strengt;
            if(forceFactor == 0f) return;
            shutDown = false;
            shutDownFactor = 4f;
            enabled = true;
        }
        
        public void ShutDown()
        {
            if(forceFactor == 0f || !enabled) return;
            shutDown = true;
        }
    }
}