using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using MeshVR;
using UnityEngine;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;

namespace CheesyFX
{
    public class AudioImporter
    {
        private static List<Request> bundles = new List<Request>();

        public static void GetClipsFromAssetBundle(List<ClipLibrary> clipLibraries, string bundlePath)
        {
            Request request = new AssetLoader.AssetBundleFromFileRequest{
                path = bundlePath,
                callback = val => OnBundleLoaded(val, clipLibraries)};
            AssetLoader.QueueLoadAssetBundleFromFile(request);
        }

        public static void GetClipsFromAssetBundle(ClipLibrary clipLibrary, string bundlePath)
        {
            var clipLibraries = new List<ClipLibrary> { clipLibrary };
            GetClipsFromAssetBundle(clipLibraries, bundlePath);
        }

        private static void OnBundleLoaded(Request request, List<ClipLibrary> clipLibraries){
            try{
                bundles.Add(request);
                ReadBundle(request, clipLibraries);
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        public static void ReadBundle(Request bundle, List<ClipLibrary> clipLibraries)
        {
            // bundle.path.Print();
            foreach (var clipLibrary in clipLibraries)
            {
                clipLibrary.Clear();
                switch (clipLibrary.name)
                {
                    case "slaps":
                    {
                        var paths = bundle.assetBundle.GetAllAssetNames().Where(x => x.Contains("/slaps/"));
                        foreach (string path in paths) {clipLibrary.AddClip(bundle.assetBundle.LoadAsset<AudioClip>(path));}
                        break;
                    }
                    case "sexslaps":
                    {
                        var paths = bundle.assetBundle.GetAllAssetNames().Where(x => x.Contains("/sexslaps"));
                        foreach (string path in paths) clipLibrary.AddClip(bundle.assetBundle.LoadAsset<AudioClip>(path));
                        break;
                    }
                    case "buttslaps":
                    {
                        var paths = bundle.assetBundle.GetAllAssetNames().Where(x => x.Contains("/buttslaps"));
                        foreach (string path in paths) clipLibrary.AddClip(bundle.assetBundle.LoadAsset<AudioClip>(path));
                        break;
                    }
                    case "squishes":
                    {
                        var paths = bundle.assetBundle.GetAllAssetNames().Where(x => x.Contains("squishes/sq0"));
                        foreach (string path in paths)
                        {
                            clipLibrary.AddClip(bundle.assetBundle.LoadAsset<AudioClip>(path));
                        }
                        break;
                    }
                    case "blowjobs":
                    {
                        var paths = bundle.assetBundle.GetAllAssetNames().Where(x => x.Contains("blowjob/f-bj0"));
                        foreach (string path in paths)
                        {
                            clipLibrary.AddClip(bundle.assetBundle.LoadAsset<AudioClip>(path));
                        }
                        break;
                    }
                }

                if (clipLibrary.name.Contains("moans"))
                {
                    string id = clipLibrary.name.Substring(5);
                    IEnumerable<string> paths;
                    if (id != "5")
                    {
                        paths = bundle.assetBundle.GetAllAssetNames().Where(x => x.Contains(ReadMyLips.voice.val.ToLower()+$"/m{id}-"));
                    }
                    else paths = bundle.assetBundle.GetAllAssetNames().Where(x => x.Contains(ReadMyLips.voice.val.ToLower()+"/o-"));
                    foreach (string path in paths)
                    {
                        clipLibrary.AddClip(bundle.assetBundle.LoadAsset<AudioClip>(path));
                    }
                }
                // clipLibrary.name.Print();
                // clipLibrary.clips.Count.Print();
            }
        }

        // public static void ReadBundle(string name, ClipLibrary clipLibrary)
        // {
        //     var clipLibraries = new List<ClipLibrary> { clipLibrary };
        //     ReadBundle(bundles.FirstOrDefault(x => x.path.Contains(name)), clipLibraries);
        // }
        
        public static IEnumerator ReadBundle(string name, List<ClipLibrary> clipLibraries)
        {
            var bundle = bundles.FirstOrDefault(x => x.path.Contains(name));
            int i = 0;
            while (bundle == null || i > 10)
            {
                yield return new WaitForSeconds(.5f);
                bundle = bundles.FirstOrDefault(x => x.path.Contains(name));
                i++;
            }
            ReadBundle(bundle, clipLibraries);
        }

        // private void GetClipLibraryFromAssetBundle(Request bundle, string library)
        // {
        //     var paths = bundle.assetBundle.GetAllAssetNames().
        //         Where(x => x.EndsWith(".ogg")||x.EndsWith(".wav")||x.EndsWith(".mp3")).
        //         Where(x => x.Split('/')[2] == library);
        //     AudioImporter.clips = paths.Select(x => bundle.assetBundle.LoadAsset<AudioClip>(x)).ToList();
        // }

        public static void UnloadBundles()
        {
            bundles.ForEach(x => AssetLoader.DoneWithAssetBundleFromFile(x.path));
        }
    }
}